/*
* Cursor2LOB v1.0.0
* Copyright (c) 2016, Michael Schmid. All rights reserved.
*
* This program is part of Cursor2LOB.
*
* Cursor2LOB is free software: you can redistribute it and/or modify
* it under the terms of the GNU Lesser General Public License as published by
* the Free Software Foundation, either version 3 of the License, or
* (at your option) any later version.
*
* Cursor2LOB is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
* GNU Lesser General Public License for more details.
*
* You should have received a copy of the GNU Lesser General Public License
* along with Cursor2LOB. If not, see <http://www.gnu.org/licenses/>.
*/
create or replace package cursor2lob
as
  invalid_charset exception;
  pragma exception_init(invalid_charset, -20000);
  --
  function to_clob(p_cursor_number    in out nocopy integer,
                   p_clob             in out nocopy clob,
                   p_row_separator    varchar2 := null,
                   p_column_separator varchar2 := null,
                   p_delimiter        varchar2 := null,
                   p_column_names     number := 0)
    return pls_integer;
  --
  function to_clob(p_cursor           in out sys_refcursor,
                   p_clob             in out nocopy clob,
                   p_row_separator    varchar2 := null,
                   p_column_separator varchar2 := null,
                   p_delimiter        varchar2 := null,
                   p_column_names     number := 0)
    return pls_integer;
  --
  procedure to_clob(p_cursor_number    in out nocopy integer,
                    p_clob             in out nocopy clob,
                    p_row_separator    varchar2 := null,
                    p_column_separator varchar2 := null,
                    p_delimiter        varchar2 := null,
                    p_column_names     number := 0);
  --
  procedure to_clob(p_cursor           in out sys_refcursor,
                    p_clob             in out nocopy clob,
                    p_row_separator    varchar2 := null,
                    p_column_separator varchar2 := null,
                    p_delimiter        varchar2 := null,
                    p_column_names     number := 0);
  --
  function to_clob(p_cursor_number    integer,
                   p_row_separator    varchar2 := null,
                   p_column_separator varchar2 := null,
                   p_delimiter        varchar2 := null,
                   p_column_names     number := 0)
    return clob;
  --
  function to_clob(p_cursor           sys_refcursor,
                   p_row_separator    varchar2 := null,
                   p_column_separator varchar2 := null,
                   p_delimiter        varchar2 := null,
                   p_column_names     number := 0)
    return clob;
  --
  function to_blob(p_cursor_number    in out nocopy integer,
                   p_blob             in out nocopy blob,
                   p_row_separator    varchar2 := null,
                   p_column_separator varchar2 := null,
                   p_character_set    varchar2 := null,
                   p_delimiter        varchar2 := null,
                   p_column_names     number := 0)
    return pls_integer;
  --
  function to_blob(p_cursor           in out sys_refcursor,
                   p_blob             in out nocopy blob,
                   p_row_separator    varchar2 := null,
                   p_column_separator varchar2 := null,
                   p_character_set    varchar2 := null,
                   p_delimiter        varchar2 := null,
                   p_column_names     number := 0)
    return pls_integer;
  --
  procedure to_blob(p_cursor_number    in out nocopy integer,
                    p_blob             in out nocopy blob,
                    p_row_separator    varchar2 := null,
                    p_column_separator varchar2 := null,
                    p_character_set    varchar2 := null,
                    p_delimiter        varchar2 := null,
                    p_column_names     number := 0);
  --
  procedure to_blob(p_cursor           in out sys_refcursor,
                    p_blob             in out nocopy blob,
                    p_row_separator    varchar2 := null,
                    p_column_separator varchar2 := null,
                    p_character_set    varchar2 := null,
                    p_delimiter        varchar2 := null,
                    p_column_names     number := 0);
  --
  function to_blob(p_cursor_number    integer,
                   p_row_separator    varchar2 := null,
                   p_column_separator varchar2 := null,
                   p_character_set    varchar2 := null,
                   p_delimiter        varchar2 := null,
                   p_column_names     number := 0)
    return blob;
  --
  function to_blob(p_cursor           sys_refcursor,
                   p_row_separator    varchar2 := null,
                   p_column_separator varchar2 := null,
                   p_character_set    varchar2 := null,
                   p_delimiter        varchar2 := null,
                   p_column_names     number := 0)
    return blob;
  --
  function to_file(p_cursor_number    in out nocopy integer,
                   p_file             utl_file.file_type,
                   p_row_separator    varchar2 := null,
                   p_column_separator varchar2 := null,
                   p_character_set    varchar2 := null,
                   p_delimiter        varchar2 := null,
                   p_column_names     number := 0)
    return pls_integer;
  --
  function to_file(p_cursor           in out sys_refcursor,
                   p_file             utl_file.file_type,
                   p_row_separator    varchar2 := null,
                   p_column_separator varchar2 := null,
                   p_character_set    varchar2 := null,
                   p_delimiter        varchar2 := null,
                   p_column_names     number := 0)
    return pls_integer;
  --
  procedure to_file(p_cursor_number    in out nocopy integer,
                    p_file             utl_file.file_type,
                    p_row_separator    varchar2 := null,
                    p_column_separator varchar2 := null,
                    p_character_set    varchar2 := null,
                    p_delimiter        varchar2 := null,
                    p_column_names     number := 0);
  --
  procedure to_file(p_cursor           in out sys_refcursor,
                    p_file             utl_file.file_type,
                    p_row_separator    varchar2 := null,
                    p_column_separator varchar2 := null,
                    p_character_set    varchar2 := null,
                    p_delimiter        varchar2 := null,
                    p_column_names     number := 0);
end;
/
show errors;
create or replace package body cursor2lob
as
  subtype lvarchar2 is varchar2(32767 byte);
  subtype lraw is raw(32767);
  --
  c_max_bytes constant pls_integer := 32767;
  --
  function get_character_set_id(p_character_set varchar2)
    return number
  as
    l_id constant number :=
      nls_charset_id(nvl(upper(trim(p_character_set)), 'CHAR_CS'));
  begin
    if l_id is null
    then
      raise_application_error(-20000, 'Invalid charset ' || p_character_set);
    else return l_id;
    end if;
  end get_character_set_id;
  --
  procedure free(p_clob in out nocopy clob) as
  begin
    if p_clob is not null and dbms_lob.istemporary(p_clob) = 1
    then
      dbms_lob.freetemporary(p_clob);
    end if;
  end free;
  --
  procedure free(p_blob in out nocopy blob) as
  begin
    if p_blob is not null and dbms_lob.istemporary(p_blob) = 1
    then
      dbms_lob.freetemporary(p_blob);
    end if;
  end free;
  --
  procedure blob2file(p_blob blob, p_file utl_file.file_type)
  as
    l_raw    lraw;
    l_amount integer := c_max_bytes;
    l_offset integer := 1;
  begin
    if p_blob is not null
    then
      begin
        loop
          dbms_lob.read(p_blob, l_amount, l_offset, l_raw);
          utl_file.put_raw(p_file, l_raw);
          l_offset := l_offset + l_amount;
          l_amount := c_max_bytes;
        end loop;
      exception
        when no_data_found then null;
      end;
      utl_file.fflush(p_file);
    end if;
  end blob2file;
  --
  function contains(p_string varchar2, p_search varchar2) return boolean as
  begin
    if p_string is null or p_search is null then return false;
    else
      for i in 1 .. length(p_search)
      loop
        if instr(p_string, substr(p_search, i, 1)) > 0 then return true; end if;
      end loop;
      return false;
    end if;
  end contains;
  --
  function contains(p_clob clob, p_search varchar2) return boolean as
  begin
    if p_clob is null or p_search is null then return false;
    else
      for i in 1 .. length(p_search)
      loop
        if instr(p_clob, substr(p_search, i, 1)) > 0 then return true; end if;
      end loop;
      return false;
    end if;
  end contains;
  --
  function escape(p_string           varchar2,
                  p_row_separator    varchar2,
                  p_column_separator varchar2,
                  p_delimiter        varchar2)
    return varchar2
  as
  begin
    if p_delimiter is not null and
       contains(p_string, p_row_separator || p_column_separator || p_delimiter)
    then
      return p_delimiter ||
             replace(p_string, p_delimiter, p_delimiter || p_delimiter) ||
             p_delimiter;
    else return p_string;
    end if;
  end escape;
  --
  function escape(p_clob             clob,
                  p_row_separator    varchar2,
                  p_column_separator varchar2,
                  p_delimiter        varchar2)
    return clob
  as
  begin
    if p_delimiter is not null and
       contains(p_clob, p_row_separator || p_column_separator || p_delimiter)
    then
      return p_delimiter ||
             replace(p_clob, p_delimiter, p_delimiter || p_delimiter) ||
             p_delimiter;
    else return p_clob;
    end if;
  end escape;
  --
  procedure close_cursor(p_cursor_number in out nocopy integer) as
  begin
    if dbms_sql.is_open(p_cursor_number)
    then
      dbms_sql.close_cursor(p_cursor_number);
    end if;
  end close_cursor;
  --
  procedure close_cursor(p_cursor        sys_refcursor,
                         p_cursor_number in out nocopy integer)
  as
  begin
    if p_cursor%isopen then close p_cursor; end if;
    --
    close_cursor(p_cursor_number);
  end close_cursor;
 
  --
  function to_clob(p_cursor_number    in out nocopy integer,
                   p_clob             in out nocopy clob,
                   p_row_separator    varchar2 := null,
                   p_column_separator varchar2 := null,
                   p_delimiter        varchar2 := null,
                   p_column_names     number := 0)
    return pls_integer
  as
    l_column_count  integer;
    l_desc          dbms_sql.desc_tab2;
    l_buffer        lvarchar2;
    l_string        lvarchar2;
    l_clob          clob;
    l_row_count     pls_integer;
    l_message       lvarchar2;
    --
    procedure append as
    begin
      if l_buffer is not null
      then
        dbms_lob.writeappend(p_clob, length(l_buffer), l_buffer);
        l_buffer := null;
      end if;
    end;
    --
    procedure append(p_string varchar2) as
    begin
      if nvl(lengthb(l_buffer), 0) + nvl(lengthb(p_string), 0) <= c_max_bytes
      then
        l_buffer := l_buffer || p_string;
      else
        dbms_lob.writeappend(p_clob, length(l_buffer), l_buffer);
        l_buffer := p_string;
      end if;
    end;
    --
    procedure append(p_append_clob clob) as
    begin append(); dbms_lob.append(p_clob, p_append_clob); end;
    --
    procedure finally as begin close_cursor(p_cursor_number); end;
  begin
    if p_cursor_number is not null
    then
      dbms_sql.describe_columns2(p_cursor_number, l_column_count, l_desc);
      for i in 1 .. l_column_count
      loop
        if l_desc(i).col_type = dbms_types.typecode_clob
        then
          dbms_sql.define_column(p_cursor_number, i, l_clob);
        else
          dbms_sql.define_column(p_cursor_number, i, l_string, c_max_bytes);
        end if;
      end loop;
      --
      while dbms_sql.fetch_rows(p_cursor_number) > 0
      loop
        l_row_count := nvl(l_row_count, 0) + 1;
        --
        if p_clob is null
        then
          dbms_lob.createtemporary(p_clob, true, dbms_lob.call);
        end if;
        --
        if p_column_names = 1 and l_row_count = 1
        then
          for i in 1 .. l_column_count
          loop
            if i > 1 then append(p_column_separator); end if;
            append(escape(l_desc(i).col_name, p_row_separator,
                          p_column_separator, p_delimiter));
          end loop;
          append(p_row_separator);
        end if;
        --
        for i in 1 .. l_column_count
        loop
          if i > 1 then append(p_column_separator); end if;
          --
          if l_desc(i).col_type = dbms_types.typecode_clob
          then
            dbms_sql.column_value(p_cursor_number, i, l_clob);
            append(escape(l_clob, p_row_separator,
                          p_column_separator, p_delimiter));
          else
            dbms_sql.column_value(p_cursor_number, i, l_string);
            append(escape(l_string, p_row_separator,
                          p_column_separator, p_delimiter));
          end if;
        end loop;
       append(p_row_separator);
      end loop;
      --
      append();
    end if;
    finally();
    return l_row_count;
  exception
    when others then finally(); raise;
  end to_clob;
  --
  function to_clob(p_cursor           in out sys_refcursor,
                   p_clob             in out nocopy clob,
                   p_row_separator    varchar2 := null,
                   p_column_separator varchar2 := null,
                   p_delimiter        varchar2 := null,
                   p_column_names     number := 0)
    return pls_integer
  as
    l_cursor_number integer;
    l_count         pls_integer;
    --
    procedure finally as begin close_cursor(p_cursor, l_cursor_number); end;
  begin
    if p_cursor is not null
    then
      l_cursor_number := dbms_sql.to_cursor_number(p_cursor);
      l_count := to_clob(p_cursor_number    => l_cursor_number,
                         p_clob             => p_clob,
                         p_row_separator    => p_row_separator,
                         p_column_separator => p_column_separator,
                         p_delimiter        => p_delimiter,
                         p_column_names     => p_column_names);
    end if;
    finally();
    return l_count;
  exception
    when others then finally(); raise;
  end to_clob;
  --
  procedure to_clob(p_cursor_number    in out nocopy integer,
                    p_clob             in out nocopy clob,
                    p_row_separator    varchar2 := null,
                    p_column_separator varchar2 := null,
                    p_delimiter        varchar2 := null,
                    p_column_names     number := 0)
  as
    l_count pls_integer;
  begin
    l_count := to_clob(p_cursor_number    => p_cursor_number,
                       p_clob             => p_clob,
                       p_row_separator    => p_row_separator,
                       p_column_separator => p_column_separator,
                       p_delimiter        => p_delimiter,
                       p_column_names     => p_column_names);
  end to_clob;
  --
  procedure to_clob(p_cursor           in out sys_refcursor,
                    p_clob             in out nocopy clob,
                    p_row_separator    varchar2 := null,
                    p_column_separator varchar2 := null,
                    p_delimiter        varchar2 := null,
                    p_column_names     number := 0)
  as
    l_count pls_integer;
  begin
    l_count := to_clob(p_cursor           => p_cursor,
                       p_clob             => p_clob,
                       p_row_separator    => p_row_separator,
                       p_column_separator => p_column_separator,
                       p_delimiter        => p_delimiter,
                       p_column_names     => p_column_names);
  end to_clob;
  --
  function to_clob(p_cursor_number    integer,
                   p_row_separator    varchar2 := null,
                   p_column_separator varchar2 := null,
                   p_delimiter        varchar2 := null,
                   p_column_names     number := 0)
    return clob
  as
    l_cursor_number integer := p_cursor_number;
    l_clob          clob;
    l_count         pls_integer;
  begin
    l_count := to_clob(p_cursor_number   => l_cursor_number,
                       p_clob             => l_clob,
                       p_row_separator    => p_row_separator,
                       p_column_separator => p_column_separator,
                       p_delimiter        => p_delimiter,
                       p_column_names     => p_column_names);
    return l_clob;
  end to_clob;
  --
  function to_clob(p_cursor           sys_refcursor,
                   p_row_separator    varchar2 := null,
                   p_column_separator varchar2 := null,
                   p_delimiter        varchar2 := null,
                   p_column_names     number := 0)
    return clob
  as
    l_cursor sys_refcursor := p_cursor;
    l_clob   clob;
    l_count  pls_integer;
  begin
    l_count := to_clob(p_cursor           => l_cursor,
                       p_clob             => l_clob,
                       p_row_separator    => p_row_separator,
                       p_column_separator => p_column_separator,
                       p_delimiter        => p_delimiter,
                       p_column_names     => p_column_names);
    return l_clob;
  end to_clob;
  --
  function to_blob(p_cursor_number    in out nocopy integer,
                   p_blob             in out nocopy blob,
                   p_row_separator    varchar2 := null,
                   p_column_separator varchar2 := null,
                   p_character_set    varchar2 := null,
                   p_delimiter        varchar2 := null,
                   p_column_names     number := 0)
    return pls_integer
  as
    l_count       pls_integer := 0;
    l_clob        clob;
    l_dest_offset integer;
    l_src_offset  integer := 1;
    l_lang_ctx    integer := dbms_lob.default_lang_ctx;
    l_warning     integer;
    --
    procedure finally as begin free(l_clob); end;
  begin
    l_count := to_clob(p_cursor_number    => p_cursor_number,
                       p_clob             => l_clob,
                       p_row_separator    => p_row_separator,
                       p_column_separator => p_column_separator,
                       p_delimiter        => p_delimiter,
                       p_column_names     => p_column_names);
    --
    if l_count > 0
    then
      if p_blob is null
      then
        dbms_lob.createtemporary(p_blob, true, dbms_lob.call);
      end if;
      --
      l_dest_offset := dbms_lob.getlength(p_blob) + 1;
      dbms_lob.converttoblob(
        dest_lob     => p_blob,
        src_clob     => l_clob,
        amount       => dbms_lob.lobmaxsize,
        dest_offset  => l_dest_offset,
        src_offset   => l_src_offset,
        blob_csid    => get_character_set_id(p_character_set),
        lang_context => l_lang_ctx,
        warning      => l_warning
      );
    end if;
    --
    finally();
    return l_count;
  exception
    when others then finally(); raise;
  end to_blob;
  --
  function to_blob(p_cursor           in out sys_refcursor,
                   p_blob             in out nocopy blob,
                   p_row_separator    varchar2 := null,
                   p_column_separator varchar2 := null,
                   p_character_set    varchar2 := null,
                   p_delimiter        varchar2 := null,
                   p_column_names     number := 0)
    return pls_integer
  as
    l_cursor_number integer;
    l_count         pls_integer;
    --
    procedure finally as begin close_cursor(p_cursor, l_cursor_number); end;
  begin
    if p_cursor is not null
    then
      l_cursor_number := dbms_sql.to_cursor_number(p_cursor);
      l_count := to_blob(p_cursor_number    => l_cursor_number,
                         p_blob             => p_blob,
                         p_row_separator    => p_row_separator,
                         p_column_separator => p_column_separator,
                         p_character_set    => p_character_set,
                         p_delimiter        => p_delimiter,
                         p_column_names     => p_column_names);
    end if;
    finally();
    return l_count;
  exception
    when others then finally(); raise;
  end to_blob;
  --
  procedure to_blob(p_cursor_number    in out nocopy integer,
                    p_blob             in out nocopy blob,
                    p_row_separator    varchar2 := null,
                    p_column_separator varchar2 := null,
                    p_character_set    varchar2 := null,
                    p_delimiter        varchar2 := null,
                    p_column_names     number := 0)
  as
    l_count pls_integer;
  begin
    l_count := to_blob(p_cursor_number    => p_cursor_number,
                       p_blob             => p_blob,
                       p_row_separator    => p_row_separator,
                       p_column_separator => p_column_separator,
                       p_character_set    => p_character_set,
                       p_delimiter        => p_delimiter,
                       p_column_names     => p_column_names);
  end to_blob;
  --
  procedure to_blob(p_cursor           in out sys_refcursor,
                    p_blob             in out nocopy blob,
                    p_row_separator    varchar2 := null,
                    p_column_separator varchar2 := null,
                    p_character_set    varchar2 := null,
                    p_delimiter        varchar2 := null,
                    p_column_names     number := 0)
  as
    l_count pls_integer;
  begin
    l_count := to_blob(p_cursor           => p_cursor,
                       p_blob             => p_blob,
                       p_row_separator    => p_row_separator,
                       p_column_separator => p_column_separator,
                       p_character_set    => p_character_set,
                       p_delimiter        => p_delimiter,
                       p_column_names     => p_column_names);
  end to_blob;
  --
  function to_blob(p_cursor_number    integer,
                   p_row_separator    varchar2 := null,
                   p_column_separator varchar2 := null,
                   p_character_set    varchar2 := null,
                   p_delimiter        varchar2 := null,
                   p_column_names     number := 0)
    return blob
  as
    l_cursor_number integer := p_cursor_number;
    l_blob          blob;
    l_count         pls_integer;
  begin
    l_count := to_blob(p_cursor_number   => l_cursor_number,
                       p_blob             => l_blob,
                       p_row_separator    => p_row_separator,
                       p_column_separator => p_column_separator,
                       p_character_set    => p_character_set,
                       p_delimiter        => p_delimiter,
                       p_column_names     => p_column_names);
    return l_blob;
  end to_blob;
  --
  function to_blob(p_cursor           sys_refcursor,
                   p_row_separator    varchar2 := null,
                   p_column_separator varchar2 := null,
                   p_character_set    varchar2 := null,
                   p_delimiter        varchar2 := null,
                   p_column_names     number := 0)
    return blob
  as
    l_cursor sys_refcursor := p_cursor;
    l_blob   blob;
    l_count  pls_integer;
  begin
    l_count := to_blob(p_cursor           => l_cursor,
                       p_blob             => l_blob,
                       p_row_separator    => p_row_separator,
                       p_column_separator => p_column_separator,
                       p_character_set    => p_character_set,
                       p_delimiter        => p_delimiter,
                       p_column_names     => p_column_names);
    return l_blob;
  end to_blob;
  --
  function to_file(p_cursor_number    in out nocopy integer,
                   p_file             utl_file.file_type,
                   p_row_separator    varchar2 := null,
                   p_column_separator varchar2 := null,
                   p_character_set    varchar2 := null,
                   p_delimiter        varchar2 := null,
                   p_column_names     number := 0)
    return pls_integer
  as
    l_blob  blob;
    l_count pls_integer;
    --
    procedure finally as begin free(l_blob); end;
  begin
    l_count := to_blob(p_cursor_number    => p_cursor_number,
                       p_blob             => l_blob,
                       p_row_separator    => p_row_separator,
                       p_column_separator => p_column_separator,
                       p_character_set    => p_character_set,
                       p_delimiter        => p_delimiter,
                       p_column_names     => p_column_names);
    --
    if l_count > 0 then blob2file(l_blob, p_file); end if;
    --
    finally();
    return l_count;
  exception
    when others then finally(); raise;
  end to_file;
  --
  function to_file(p_cursor           in out sys_refcursor,
                   p_file             utl_file.file_type,
                   p_row_separator    varchar2 := null,
                   p_column_separator varchar2 := null,
                   p_character_set    varchar2 := null,
                   p_delimiter        varchar2 := null,
                   p_column_names     number := 0)
    return pls_integer
  as
    l_blob  blob;
    l_count pls_integer;
    --
    procedure finally as begin free(l_blob); end;
  begin
    l_count := to_blob(p_cursor           => p_cursor,
                       p_blob             => l_blob,
                       p_row_separator    => p_row_separator,
                       p_column_separator => p_column_separator,
                       p_character_set    => p_character_set,
                       p_delimiter        => p_delimiter,
                       p_column_names     => p_column_names);
    --
    if l_count > 0 then blob2file(l_blob, p_file); end if;
    --
    finally();
    return l_count;
  exception
    when others then finally(); raise;
  end to_file;
  --
  procedure to_file(p_cursor_number    in out nocopy integer,
                    p_file             utl_file.file_type,
                    p_row_separator    varchar2 := null,
                    p_column_separator varchar2 := null,
                    p_character_set    varchar2 := null,
                    p_delimiter        varchar2 := null,
                    p_column_names     number := 0)
  as
    l_count pls_integer;
  begin
    l_count := to_file(p_cursor_number    => p_cursor_number,
                       p_file             => p_file,
                       p_row_separator    => p_row_separator,
                       p_column_separator => p_column_separator,
                       p_character_set    => p_character_set,
                       p_delimiter        => p_delimiter,
                       p_column_names     => p_column_names);
  end to_file;
  --
  procedure to_file(p_cursor           in out sys_refcursor,
                    p_file             utl_file.file_type,
                    p_row_separator    varchar2 := null,
                    p_column_separator varchar2 := null,
                    p_character_set    varchar2 := null,
                    p_delimiter        varchar2 := null,
                    p_column_names     number := 0)
  as
    l_count pls_integer;
  begin
    l_count := to_file(p_cursor           => p_cursor,
                       p_file             => p_file,
                       p_row_separator    => p_row_separator,
                       p_column_separator => p_column_separator,
                       p_character_set    => p_character_set,
                       p_delimiter        => p_delimiter,
                       p_column_names     => p_column_names);
  end to_file;
end;
/
show errors;
